/*
 * glshader gstreamer plugin
 * Copyrithg (C) 2008 Filippo Argiolas <filippo.argiolas@gmail.com>
 * Copyright (C) 2009 Luc Deschenaux <luc.deschenaux@freesurf.ch>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-glshader
 *
 * Filter loading OpenGL fragment shader from file
 *
 * <refsect2>
 * <title>Examples</title>
 * |[
 * gst-launch videotestsrc ! glupload ! glshader location=myshader.fs ! glimagesink
 * ]|
 * FBO (Frame Buffer Object) and GLSL (OpenGL Shading Language) are required.
 * </refsect2>
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gst/gst.h>
#include <gst/controller/gstcontroller.h>
#include "gstglfiltershader.h"
#include <gstglshadervariables.h>

/* horizontal filter */
static gchar *hfilter_fragment_source;
static gchar *hfilter_fragment_variables[2];

enum {
	PROP_0,
	PROP_LOCATION,
	PROP_PRESET,
	PROP_VARIABLES
};

#define GST_CAT_DEFAULT gst_gl_filtershader_debug
GST_DEBUG_CATEGORY_STATIC (GST_CAT_DEFAULT);

static const GstElementDetails element_details =
GST_ELEMENT_DETAILS ("OpenGL fragment shader filter",
    "Filter/Effect",
    "Filter loading OpenGL fragment shader from file",
    "<luc.deschenaux@freesurf.ch>, Filippo Argiolas <filippo.argiolas@gmail.com>");

#define DEBUG_INIT(bla)							\
  GST_DEBUG_CATEGORY_INIT (gst_gl_filtershader_debug, "glshader", 0, "glshader element");

GST_BOILERPLATE_FULL (GstGLFilterShader, gst_gl_filtershader, GstGLFilter,
    GST_TYPE_GL_FILTER, DEBUG_INIT);

static void gst_gl_filtershader_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_gl_filtershader_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static void gst_gl_filter_filtershader_reset (GstGLFilter * filter);
static void gst_gl_filtershader_draw_texture (GstGLFilterShader * filtershader,
    GLuint tex);

static void gst_gl_filtershader_load_shader (char *filename);
static void gst_gl_filtershader_load_variables (char *filename);
static void gst_gl_filtershader_init_shader (GstGLFilter * filter);
static gboolean gst_gl_filtershader_filter (GstGLFilter * filter,
    GstGLBuffer * inbuf, GstGLBuffer * outbuf);
static void gst_gl_filtershader_hcallback (gint width, gint height, guint texture,
    gpointer stuff);


static void
gst_gl_filtershader_init_resources (GstGLFilter * filter)
{
 // GstGLFilterShader *filtershader = GST_GL_FILTERSHADER (filter);

///  glGenTextures (1, &filtershader->midtexture);
///  glBindTexture (GL_TEXTURE_RECTANGLE_ARB, filtershader->midtexture);
  glTexImage2D (GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA8,
      filter->width, filter->height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
  glTexParameteri (GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri (GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri (GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S,
      GL_CLAMP_TO_EDGE);
  glTexParameteri (GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T,
      GL_CLAMP_TO_EDGE);
}

static void
gst_gl_filtershader_reset_resources (GstGLFilter * filter)
{
  //GstGLFilterShader *filtershader = GST_GL_FILTERSHADER (filter);

///  glDeleteTextures (1, &filtershader->midtexture);
}

static void
gst_gl_filtershader_base_init (gpointer klass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (klass);

  gst_element_class_set_details (element_class, &element_details);
  /* initialize library */
  gst_controller_init (NULL, NULL);

}

static void
gst_gl_filtershader_class_init (GstGLFilterShaderClass * klass)
{
  GObjectClass *gobject_class;

  gobject_class = (GObjectClass *) klass;
  gobject_class->set_property = gst_gl_filtershader_set_property;
  gobject_class->get_property = gst_gl_filtershader_get_property;

  g_object_class_install_property (gobject_class, PROP_LOCATION,
      g_param_spec_string ("location", "File Location",
          "Location of the GLSL file to load", NULL,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  g_object_class_install_property (gobject_class, PROP_PRESET,
      g_param_spec_string ("preset", "Preset File Location",
          "Location of the shader uniform variables preset file", NULL,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  g_object_class_install_property (gobject_class, PROP_VARIABLES,
      g_param_spec_string ("vars", "Uniform variables",
          "Set the shader uniform variables", NULL,
          G_PARAM_READWRITE | GST_PARAM_CONTROLLABLE | G_PARAM_STATIC_STRINGS));

  GST_GL_FILTER_CLASS (klass)->filter = gst_gl_filtershader_filter;
  GST_GL_FILTER_CLASS (klass)->display_init_cb =
      gst_gl_filtershader_init_resources;
  GST_GL_FILTER_CLASS (klass)->display_reset_cb =
      gst_gl_filtershader_reset_resources;
  GST_GL_FILTER_CLASS (klass)->onInitFBO = gst_gl_filtershader_init_shader;
  GST_GL_FILTER_CLASS (klass)->onReset = gst_gl_filter_filtershader_reset;
}

static void
gst_gl_filtershader_init (GstGLFilterShader * filtershader,
    GstGLFilterShaderClass * klass)
{
  filtershader->shader0 = NULL;
///  filtershader->shader1 = NULL;
///  filtershader->midtexture = 0;
}

static void
gst_gl_filter_filtershader_reset (GstGLFilter * filter)
{
  GstGLFilterShader *filtershader = GST_GL_FILTERSHADER (filter);

  //blocking call, wait the opengl thread has destroyed the shader
  gst_gl_display_del_shader (filter->display, filtershader->shader0);

  //blocking call, wait the opengl thread has destroyed the shader
///  gst_gl_display_del_shader (filter->display, filtershader->shader1);
}

static void
gst_gl_filtershader_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstGLFilterShader *filtershader = GST_GL_FILTERSHADER (object);

  switch (prop_id) {

    case PROP_LOCATION:

      if (filtershader->filename) {
	      g_free(filtershader->filename);
      }
      filtershader->filename=g_strdup(g_value_get_string (value));

      break;

    case PROP_PRESET:

      if (filtershader->presetfile) {
	      g_free(filtershader->presetfile);
      }
      filtershader->presetfile=g_strdup(g_value_get_string (value));

      break;

    case PROP_VARIABLES:

      if (hfilter_fragment_variables[1]) {
	      gst_gl_shadervariables_cache_delete(hfilter_fragment_variables[1]);
	      g_free(hfilter_fragment_variables[1]);
      }
      hfilter_fragment_variables[1]=g_strdup(g_value_get_string (value));

      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_gl_filtershader_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
   GstGLFilterShader *filtershader = GST_GL_FILTERSHADER (object); 

  switch (prop_id) {
    case PROP_LOCATION:
      g_value_set_string(value, filtershader->filename);
      break;

    case PROP_PRESET:
      g_value_set_string(value, filtershader->presetfile);
      break;

    case PROP_VARIABLES:
      g_value_set_string(value, hfilter_fragment_variables[1]);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static int
gst_gl_filtershader_load_file (char *filename,char **storage)
{

  size_t count;
  ssize_t bytes;
  int f;

  // read the filter from file
  GST_INFO ("loading file: %s", filename);
  f=open(filename,O_RDONLY);
  if (f<0) {
	GST_ERROR ("could not open file: %s", filename);
	return -1;
  }

  if (*storage) {
	g_free(*storage);
	*storage=0;
  }

  count=lseek(f,0,SEEK_END);
  *storage=g_malloc(count+1);
  if (!*storage) {
	GST_ERROR ("g_malloc failed: %u", count);
	return -1;
  }

  lseek(f,0,SEEK_SET);
  bytes=read(f,(void*)*storage,count);
  if (bytes<0 || bytes!=count) {
	GST_ERROR ("read failed: %u/%u", bytes,count);
	return -1;
  }
  ((char*)*storage)[count]=0;

  close(f);
  GST_INFO ("read: %u", bytes);
  return 0;
}

static void
gst_gl_filtershader_load_shader (char *filename) {
	if (gst_gl_filtershader_load_file(filename,&hfilter_fragment_source)) {
		exit(1);
	}
}

static void
gst_gl_filtershader_load_variables (char *filename) {

	if (!filename) return;
	
	if (hfilter_fragment_variables[0]) {
		gst_gl_shadervariables_cache_delete(hfilter_fragment_variables[0]);
		g_free(hfilter_fragment_variables[1]);
	}

	if (gst_gl_filtershader_load_file(filename,&hfilter_fragment_variables[0])) {
		exit(1);
	}
}

static void
gst_gl_filtershader_variables_parse(GstGLShader *shader,gchar *variables)
{
	gst_gl_shadervariables_parse(shader,variables);
}


static void
gst_gl_filtershader_init_shader (GstGLFilter * filter)
{

  GstGLFilterShader *filtershader = GST_GL_FILTERSHADER (filter);

  gst_gl_filtershader_load_shader(filtershader->filename);
  gst_gl_filtershader_load_variables(filtershader->presetfile);

  //blocking call, wait the opengl thread has compiled the shader
  gst_gl_display_gen_shader (filter->display, 0, hfilter_fragment_source,
      &filtershader->shader0);

  if (!filtershader->shader0) {
	  GST_ERROR("Fragment Shader compilation failed.");
	  exit(1);
  }

  gst_gl_filtershader_variables_parse(filtershader->shader0,hfilter_fragment_variables[0]);
  gst_gl_filtershader_variables_parse(filtershader->shader0,hfilter_fragment_variables[1]);

  //blocking call, wait the opengl thread has compiled the shader
///  gst_gl_display_gen_shader (filter->display, 0, vfilter_fragment_source,
///      &filtershader->shader1);
}

static void
gst_gl_filtershader_draw_texture (GstGLFilterShader * filtershader, GLuint tex)
{
  GstGLFilter *filter = GST_GL_FILTER (filtershader);

  glActiveTexture (GL_TEXTURE0);
  glEnable (GL_TEXTURE_RECTANGLE_ARB);
  glBindTexture (GL_TEXTURE_RECTANGLE_ARB, tex);

  glBegin (GL_QUADS);

  glTexCoord2f (0.0, 0.0);
  glVertex2f (-1.0, -1.0);
  glTexCoord2f ((gfloat) filter->width, 0.0);
  glVertex2f (1.0, -1.0);
  glTexCoord2f ((gfloat) filter->width, (gfloat) filter->height);
  glVertex2f (1.0, 1.0);
  glTexCoord2f (0.0, (gfloat) filter->height);
  glVertex2f (-1.0, 1.0);

  glEnd ();
}

///static void
///change_view (GstGLDisplay * display, gpointer data)
///{
//  GstGLFilterShader *filtershader = GST_GL_FILTERSHADER (data);

///  const double mirrormatrix[16] = {
///    -1.0, 0.0, 0.0, 0.0,
///    0.0, 1.0, 0.0, 0.0,
///    0.0, 0.0, 1.0, 0.0,
///    0.0, 0.0, 0.0, 1.0
///  };

///  glMatrixMode (GL_MODELVIEW);
///  glLoadMatrixd (mirrormatrix);
///}

static gboolean
gst_gl_filtershader_filter (GstGLFilter * filter, GstGLBuffer * inbuf,
    GstGLBuffer * outbuf)
{
  GstClockTime timestamp;
 
  GstGLFilterShader *filtershader = GST_GL_FILTERSHADER (filter);

  gst_gl_filter_render_to_target (filter, inbuf->texture,  ///
      outbuf->texture, gst_gl_filtershader_hcallback, filtershader); ///

  timestamp=GST_BUFFER_TIMESTAMP (outbuf);
  if (GST_CLOCK_TIME_IS_VALID (timestamp))
      gst_object_sync_values (G_OBJECT (filter), timestamp);

///  gst_gl_filter_render_to_target (filter, inbuf->texture,
///      filtershader->midtexture, gst_gl_filtershader_hcallback, filtershader);

/// gst_gl_display_thread_add (filter->display, change_view, filtershader);

///  gst_gl_filter_render_to_target (filter, filtershader->midtexture,
///      outbuf->texture, gst_gl_filtershader_vcallback, filtershader);

  return TRUE;
}

static void
gst_gl_filtershader_hcallback (gint width, gint height, guint texture,
    gpointer stuff)
{
  GstGLFilterShader *filtershader = GST_GL_FILTERSHADER (stuff);

//  /* hard coded kernel, it could be easily generated at runtime with a
//   * property to change standard deviation */
//  gfloat gauss_kernel[9] = {
//    0.026995f, 0.064759f, 0.120985f,
//    0.176033f, 0.199471f, 0.176033f,
//    0.120985f, 0.064759f, 0.026995f
//  };

  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();

  gst_gl_shader_use (filtershader->shader0);

  glActiveTexture (GL_TEXTURE1);
  glEnable (GL_TEXTURE_RECTANGLE_ARB);
  glBindTexture (GL_TEXTURE_RECTANGLE_ARB, texture);
  glDisable (GL_TEXTURE_RECTANGLE_ARB);

  gst_gl_shader_set_uniform_1i (filtershader->shader0, "tex", 1);
  gst_gl_filtershader_variables_parse(filtershader->shader0,hfilter_fragment_variables[0]);
  gst_gl_filtershader_variables_parse(filtershader->shader0,hfilter_fragment_variables[1]);


//  gst_gl_shader_set_uniform_1fv (filtershader->shader0, "kernel", 9,
//      gauss_kernel);
//  gst_gl_shader_set_uniform_1f (filtershader->shader0, "norm_const", 0.977016f);
//  gst_gl_shader_set_uniform_1f (filtershader->shader0, "norm_offset", 0.0f);

  gst_gl_filtershader_draw_texture (filtershader, texture);
}


///static void
///gst_gl_filtershader_vcallback (gint width, gint height, guint texture,
///    gpointer stuff)
///{
///  GstGLFilterShader *filtershader = GST_GL_FILTERSHADER (stuff);

// /* hard coded kernel, it could be easily generated at runtime with a
//   * property to change standard deviation */
//  gfloat gauss_kernel[9] = {
//    0.026995f, 0.064759f, 0.120985f,
//    0.176033f, 0.199471f, 0.176033f,
//    0.120985f, 0.064759f, 0.026995f
//  };

///  glMatrixMode (GL_PROJECTION);
///  glLoadIdentity ();

///  gst_gl_shader_use (filtershader->shader1);

///  glActiveTexture (GL_TEXTURE1);
///  glEnable (GL_TEXTURE_RECTANGLE_ARB);
///  glBindTexture (GL_TEXTURE_RECTANGLE_ARB, texture);
///  glDisable (GL_TEXTURE_RECTANGLE_ARB);

///  gst_gl_shader_set_uniform_1i (filtershader->shader1, "tex", 1);

//  gst_gl_shader_set_uniform_1fv (filtershader->shader1, "kernel", 9,
//      gauss_kernel);
//  gst_gl_shader_set_uniform_1f (filtershader->shader1, "norm_const", 0.977016f);
//  gst_gl_shader_set_uniform_1f (filtershader->shader1, "norm_offset", 0.0f);

///  gst_gl_filtershader_draw_texture (filtershader, texture);
///}

